import logging
from typing import List, Optional

import docker
from docker.errors import APIError, ImageNotFound, NotFound
from docker.models.containers import Container
from docker.models.images import Image


def docker_service_available():
    try:
        docker.from_env()
        return True
    except Exception:
        return False


class DockerClient:
    def __init__(self):
        try:
            self.client = docker.from_env()
        except Exception as e:
            logging.error(f"An error occurred when trying to connect to Docker: {e}")
            logging.error(
                "This error usually occurs when Docker is not running or installed.\n"
                "You can install Docker Engine from https://docs.docker.com/engine/install/\n"
            )

    def list_containers(self) -> List[Container]:
        """
        List all Docker containers with the launchflow_managed label.
        """
        try:
            # lists all containers with the launchflow_managed label
            containers = self.client.containers.list(
                filters={"label": "launchflow_managed"}
            )
            logging.info(f"Found {len(containers)} containers.")
            return containers
        except APIError as e:
            logging.error(f"Server returned error when trying to list containers: {e}")
            return []
        except Exception as e:
            logging.error(f"An error occurred when trying to list containers: {e}")
            return []

    def get_container(self, name: str) -> Optional[Container]:
        """
        Get a Docker container by name.
        """
        try:
            container = self.client.containers.get(name)
            logging.info(f"Container {name} found.")
            return container
        except NotFound:
            logging.warning(f"No container with name {name} found.")
            return None
        except APIError as e:
            logging.error(
                f"Server returned error when trying to get container {name}: {e}"
            )
            return None

    def start_container(
        self,
        name: str,
        image: str,
        env_vars=None,
        ports=None,
        volumes=None,
        **kwargs,
    ) -> Container:
        """
        Start a Docker container from an image with a given name, environment variables, ports, and volumes.
        """
        try:
            logging.info(
                f"Attempting to start container '{name}' using image '{image}'."
            )
            container = self.client.containers.run(
                image,
                name=name,
                detach=True,
                environment=env_vars,
                ports=ports,
                volumes=volumes,
                labels={"launchflow_managed": "true"},
                **kwargs,
            )
            logging.info(f"Container '{name}' started successfully.")
            return container
        except ImageNotFound:
            logging.warning(f"Image '{image}' not found. Attempting to pull image.")
            self.pull_image(image)
            return self.start_container(name, image, env_vars, ports, volumes, **kwargs)
        except APIError as e:
            logging.error(
                f"Server returned error when trying to start container '{name}': {e}"
            )
            return None

    def stop_container(self, name: str):
        """
        Stop a Docker container by name.
        """
        try:
            container = self.get_container(name)
            if container:
                logging.info(f"Stopping container '{name}'.")
                container.stop()
                logging.info(f"Container '{name}' stopped successfully.")
        except NotFound:
            logging.warning(f"No container with name '{name}' to stop.")
        except APIError as e:
            logging.error(
                f"Server returned error when trying to stop container '{name}': {e}"
            )

    def remove_container(self, name: str, force=False):
        """
        Remove a Docker container by name. Can be forced to remove running containers.
        """
        try:
            container = self.get_container(name)
            if container:
                logging.info(f"Removing container '{name}'.")
                container.remove(force=force)
                logging.info(f"Container '{name}' removed successfully.")
        except NotFound:
            logging.warning(f"No container with name '{name}' to remove.")
        except APIError as e:
            logging.error(
                f"Server returned error when trying to remove container '{name}': {e}"
            )

    def pull_image(self, image: str) -> Image:
        """
        Pull a Docker image by name.
        """
        try:
            pulled_image = self.client.images.pull(image)
            logging.info(f"Image '{image}' pulled successfully.")
            return pulled_image
        except APIError as e:
            logging.error(
                f"Server returned error when trying to pull image '{image}': {e}"
            )
            return None

    def build_and_push_image(self, config):
        """
        Build and push a Docker image according to the provided config.
        This function is not fully implemented as it requires additional config details.
        """
        logging.warning("The build_and_push_image method is not yet implemented.")
        pass


# Example usage of DockerClient
if __name__ == "__main__":
    docker_client = DockerClient()

    env_vars = {
        "POSTGRES_USER": "user",
        "POSTGRES_PASSWORD": "password",
        "POSTGRES_DB": "mydb",
    }
    port_bindings = {"5432/tcp": 5433}  # Port mapping from container to host
    volumes = {
        "/local/path/to/data": {"bind": "/var/lib/postgresql/data", "mode": "rw"}
    }

    # docker_client.stop_container("my_postgres_container")

    # docker_client.remove_container("my_postgres_container")

    container = docker_client.start_container(
        name="my_postgres_container",
        image="postgres",
        env_vars=env_vars,
        ports=port_bindings,
        volumes=volumes,
    )

    if container:
        print(f"Started Postgres container with ID: {container.id}")

    print(docker_client.list_containers())
